package DataFlow;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;

import CFG.CFG;
import CFG.ReverseCFG;

/**
 * Perform a dataflow analysis on an AST
 * @author matze
 */
public class DataFlow<Data, CFGNode> {
	
	/** data at entries of blocks */
	private Map<CFGNode, Data> data_entry = new HashMap<CFGNode, Data>();
	/** data at ext of blocks */
	private Map<CFGNode, Data> data_exit = new HashMap<CFGNode, Data>();
	/** instance of a dataflow analysis */
	private DataFlowInterface<Data, CFGNode> analysis;
		
	public DataFlow(DataFlowInterface<Data, CFGNode> analysis) {
		this.analysis = analysis;
	}

	/** return data at entry of a block */
	public Data getEntry(CFGNode node) {
		return data_entry.get(node);
	}
	
	/** return data at exit of block */
	public Data getExit(CFGNode node) {
		return data_exit.get(node);
	}
	
	/** (re)set data at block entry */
	private void setEntry(CFGNode node, Data data) {
		data_entry.put(node, data);
	}
	
	/** (re)set data at block exit */
	private void setExit(CFGNode node, Data data) {
		data_exit.put(node, data);
	}
	
	/** perform data flow analysis on a cfg */
	public void perform(CFG<CFGNode> initial_cfg) {
		/* initialize dataflow analysis */
		analysis.initialize(initial_cfg);
		
		/* reverse CFG if we have a backwards analysis */
		CFG<CFGNode> cfg = initial_cfg;
		if (analysis.isBackward())
			cfg = new ReverseCFG<CFGNode>(cfg);
		
		/* create worklist */
		Queue<CFGNode> worklist = new LinkedList<CFGNode>();
		worklist.add(cfg.getStart());
		
		while (!worklist.isEmpty()) {
			CFGNode node = worklist.poll();
			
			/* determine data at block entry */
			Data data = null;
			for (CFGNode pred : cfg.getPreds(node)) {
				Data predData = getExit(pred);
				if (predData == null)
					continue;
				
				if (data == null)
					data = predData;
				else
					data = analysis.union(data, predData);
			}
			setEntry(node, data);
			
			/* apply transfer function */
			Data oldData = getExit(node);
			Data newData = analysis.transfer(node, data);
			/* no changes? continue */
			if (newData.equals(oldData)) {
				continue;
			}
			setExit(node, newData);
			
			/* enqueue successors */
			for (CFGNode succ : cfg.getSuccs(node)) {
				worklist.add(succ);
			}
		}
		
		/* for backwards analysis the meaning of entry/exit changed */
		if (analysis.isBackward()) {
			Map<CFGNode, Data> temp = data_exit;
			data_exit = data_entry;
			data_entry = temp;
		}
	}
}
